#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Console script for ifr."""

import logging
import sys
from typing import Callable

import click

from spsdk.apps.pfr import _parse_binary_data, _store_output
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    FC,
    CommandsTreeGroup,
    spsdk_apps_common_options,
    spsdk_config_option,
    spsdk_family_option,
    spsdk_mboot_interface,
    spsdk_output_option,
    spsdk_revision_option,
)
from spsdk.apps.utils.utils import catch_spsdk_error, deprecated_option_warning, format_raw_data
from spsdk.exceptions import SPSDKError
from spsdk.mboot.mcuboot import McuBoot
from spsdk.mboot.protocol.base import MbootProtocolBase
from spsdk.pfr import pfr
from spsdk.utils.misc import load_binary, load_configuration, size_fmt, write_file
from spsdk.utils.schema_validator import CommentedConfig

logger = logging.getLogger(__name__)


def ifr_device_type_options(no_type: bool = False) -> Callable:
    """IFR common device click options.

    :param no_type: If true, the type option is not added, defaults to False
    :return: Click decorator
    """

    def decorator(options: Callable[[FC], FC]) -> Callable[[FC], FC]:
        """Setup IFR options for device, revision and sector."""
        options = click.option(
            "-r",
            "--revision",
            help="Chip revision; if not specified, most recent one will be used",
            default="latest",
        )(options)
        options = click.option(
            "-f",
            "--family",
            type=click.Choice(pfr.ROMCFG.get_supported_families(), case_sensitive=False),
            help="Device to use",
            required=True,
        )(options)
        if not no_type:
            options = click.option(
                "-s",
                "--sector",
                "sector",
                required=False,
                type=click.Choice(["ROMCFG", "CMACTable"], case_sensitive=False),
                default="ROMCFG",
                help="Select IFR sector",
            )(options)
        return options

    return decorator


@click.group(name="ifr", no_args_is_help=True, cls=CommandsTreeGroup)
@spsdk_apps_common_options
def main(log_level: int) -> int:
    """Utility for generating and parsing IFR. Please note that IFR0 ROMCFG region is one-time-programmable only."""
    spsdk_logger.install(level=log_level)
    return 0


@main.command(name="get-template", no_args_is_help=True)
@spsdk_output_option(
    required=False,
    force=True,
    help="Save the output into a file instead of console",
)
@ifr_device_type_options()
@click.option("--full", is_flag=True, help="Show full config, including computed values")
def get_template(family: str, revision: str, sector: str, output: str, full: bool) -> None:
    """Generate user configuration template file."""
    if full:
        deprecated_option_warning("full")
    ifr_cls = pfr.get_ifr_pfr_class(sector, family)
    schemas = ifr_cls.get_validation_schemas(family=family, revision=revision)
    yaml_data = CommentedConfig("IFR configuration template", schemas).get_template()
    _store_output(yaml_data, output, msg="IFR configuration template has been created.")


@main.command(name="parse-binary", no_args_is_help=True)
@ifr_device_type_options()
@spsdk_output_option(
    required=False,
    help="Save the output into a file instead of console",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, resolve_path=True),
    required=True,
    help="Binary to parse",
)
@click.option("-d", "--show-diff", is_flag=True, help="Show differences comparing to defaults")
def parse_binary(
    revision: str, output: str, binary: str, sector: str, show_diff: bool, family: str
) -> None:
    """Parse binary and extract configuration."""
    yaml_data = _parse_binary_data(
        data=load_binary(binary),
        family=family,
        revision=revision,
        area=sector,
        show_diff=show_diff,
    )
    _store_output(yaml_data, output, msg=f"Success. (IFR: {binary} has been parsed.")


@main.command(name="generate-binary", no_args_is_help=True)
@spsdk_family_option(families=pfr.ROMCFG.get_supported_families())
@spsdk_revision_option
@spsdk_config_option()
@spsdk_output_option(
    required=False,
    help="Save the output into a file instead of console",
)
def generate_binary(output: str, config: str, family: str, revision: str) -> None:
    """Generate binary data."""
    if family:
        deprecated_option_warning("family")
    if revision:
        deprecated_option_warning("revision")
    ifr_config = load_configuration(str(config))
    pfr.ROMCFG.validate_config(ifr_config)
    ifr_obj = pfr.ROMCFG.load_from_config(ifr_config)
    data = ifr_obj.export()
    _store_output(data, output, "wb", msg="Success. (IFR binary has been generated)")


@main.command(name="write", no_args_is_help=True)
@spsdk_mboot_interface(use_long_timeout_form=True, identify_by_family=True)
@spsdk_family_option(families=pfr.ROMCFG.get_supported_families())
@spsdk_revision_option
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
    help="Path to IFR data to write.",
)
def write(
    interface: MbootProtocolBase,
    family: str,
    revision: str,
    binary: str,
) -> None:
    """Write IFR page to the device."""
    ifr_obj = pfr.ROMCFG(family=family, revision=revision)
    ifr_page_address = ifr_obj.db.get_int(ifr_obj.FEATURE_NAME, ifr_obj.DB_SUB_KEYS + ["address"])
    ifr_page_length = ifr_obj.BINARY_SIZE

    click.echo(f"{ifr_obj.__class__.__name__} page address on {family} is {ifr_page_address:#x}")

    data = load_binary(binary)
    if len(data) != ifr_page_length:
        raise SPSDKError(
            f"IFR page length is {ifr_page_length}. Provided binary has {size_fmt(len(data))}."
        )

    with McuBoot(interface=interface) as mboot:
        mboot.write_memory(address=ifr_page_address, data=data)
    click.echo(
        f"{ifr_obj.__class__.__name__} data {'written to device.' if mboot.status_code == 0 else 'write failed!'}"
    )


@main.command(name="read", no_args_is_help=True)
@spsdk_mboot_interface(use_long_timeout_form=True)
@ifr_device_type_options()
@spsdk_output_option(
    required=False,
    help="Store IFR data into a file. If not specified hexdump data into stdout.",
)
@click.option(
    "-y",
    "--yaml",
    "yaml_output",
    type=click.Path(dir_okay=False, resolve_path=True),
    help="Parse data read from device into YAML config.",
)
@click.option(
    "-d",
    "--show-diff",
    is_flag=True,
    help="(applicable for parsing) Show differences comparing to defaults",
)
@click.option(
    "-c",
    "--show-calc",
    is_flag=True,
    hidden=True,
    help="(applicable for parsing) Show also calculated fields when displaying difference to "
    "defaults (--show-diff)",
)
def read(
    interface: MbootProtocolBase,
    family: str,
    revision: str,
    sector: str,
    output: str,
    yaml_output: str,
    show_diff: bool,
    show_calc: bool,
) -> None:
    """Read IFR page from the device."""
    if show_calc:
        logger.warning(
            "Show calculated fields is obsolete function for configuration YAML files."
            " In case of debugging those values check the binary data."
        )
    ifr_obj = pfr.ROMCFG(family=family, revision=revision)
    ifr_page_address = ifr_obj.db.get_int(ifr_obj.FEATURE_NAME, ifr_obj.DB_SUB_KEYS + ["address"])
    ifr_page_length = ifr_obj.BINARY_SIZE
    ifr_page_name = ifr_obj.__class__.__name__

    click.echo(f"{ifr_page_name} page address on {family} is {ifr_page_address:#x}")

    with McuBoot(interface=interface) as mboot:
        data = mboot.read_memory(address=ifr_page_address, length=ifr_page_length)
    if not data:
        raise SPSDKError(f"Unable to read data from address {ifr_page_address:#x}")

    if output:
        write_file(data, output, "wb")
        click.echo(f"{ifr_page_name} data stored to {output}")
    if yaml_output:
        yaml_data = _parse_binary_data(
            data=data,
            family=family,
            revision=revision,
            area=sector,
            show_diff=show_diff,
        )
        write_file(yaml_data, yaml_output)
        click.echo(f"Parsed config stored to {yaml_output}")
    if not output and not yaml_output:
        click.echo(format_raw_data(data=data, use_hexdump=True))


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pragma: no cover  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()  # pragma: no cover
